from typing import List
import pandas as pd


class JsonLDConverter:
    """
    Custom convertor from ORKG JSON to JSON-LD
    """

    _blank_node_counter = 1

    def __init__(self, host: str):
        self.ORKG_PREFIX = host
        if self.ORKG_PREFIX[-1] != '/':
            self.ORKG_PREFIX += '/'
        self.CLASS_PREFIX = f"{self.ORKG_PREFIX}class/"
        self.RESOURCE_PREFIX = f"{self.ORKG_PREFIX}resource/"
        self.PREDICATE_PREFIX = f"{self.ORKG_PREFIX}property/"

    def __handle_csvw_dataset(self, lol: List[List], context: dict) -> dict:
        # following CSVW @ https://www.w3.org/TR/2015/REC-tabular-data-model-20151217/
        if len(lol) == 0:
            return {}
        # create table node - the node that will be returned
        table_node = {'@type': [f'{self.CLASS_PREFIX}Table'], 'label': 'Table'}
        # create columns and assign them
        columns = lol[0]
        context['columns'] = f'{self.PREDICATE_PREFIX}CSVW_Columns'
        context['name'] = f'{self.PREDICATE_PREFIX}CSVW_Name'
        context['number'] = f'{self.PREDICATE_PREFIX}CSVW_Number'
        column_nodes = [{'@type': [f'{self.CLASS_PREFIX}Column'], 'name': c, 'number': i, '@id': self.get_blank_node_id()} for i, c in enumerate(columns, start=1)]
        table_node['columns'] = column_nodes
        # iterate over rows and create them
        row_nodes = []
        for i, r in enumerate(lol[1:], start=1):
            row = {'@type': [f'{self.CLASS_PREFIX}Row'], 'titles': f'Row {i}', 'label': f'Row {i}', 'cells': [], 'number': i}
            # run through cells
            for idx, c in enumerate(r):
                row['cells'].append({'@type': [f'{self.CLASS_PREFIX}Cell'], 'value': str(c) if not pd.isna(c) else None, 'column': column_nodes[idx]['@id']})
            # append row to rows
            row_nodes.append(row)
        context['rows'] = f'{self.PREDICATE_PREFIX}CSVW_Rows'
        context['cells'] = f'{self.PREDICATE_PREFIX}CSVW_Cells'
        context['value'] = f'{self.PREDICATE_PREFIX}CSVW_Value'
        context['titles'] = f'{self.PREDICATE_PREFIX}CSVW_Titles'
        context['column'] = f'{self.PREDICATE_PREFIX}CSVW_Column'
        # context['column'] = {"@id": f'{self.PREDICATE_PREFIX}CSVW_Column', "@type": "@id"}
        table_node['rows'] = row_nodes
        return table_node

    def get_blank_node_id(self):
        """
        Creates an incremental blank node identifier for the json-ld structure
        :return: string in the form of _:n{number}
        """
        bn_id = f"_:n{self._blank_node_counter}"
        self._blank_node_counter = self._blank_node_counter + 1
        return bn_id

    def convert_2_jsonld(self, orkg_json: dict):
        """
        Convert an ORKG JSON format to an equivalent JSON-LD format
        :param orkg_json: The root ORKG JSON object
        :return: a json object in the format of JSON-LD
        """
        context = {}
        jsonld = {}
        if 'resource' in orkg_json:
            orkg_json = orkg_json['resource']
        # create dummy blank node
        jsonld['@id'] = self.get_blank_node_id()

        # handle label
        jsonld['label'] = orkg_json['name']
        context['label'] = "http://www.w3.org/2000/01/rdf-schema#label"

        # handle classes
        jsonld['@type'] = [self.CLASS_PREFIX + c for c in orkg_json['classes']]

        # handle values
        self.handle_values(orkg_json['values'], jsonld, context)

        # prepare final jsonld
        jsonld["@context"] = context
        return jsonld

    def handle_values(self, orkg_json: dict, jsonld: dict, context: dict):
        """
        Recursively handle the values part of the ORKG JSON format
        :param orkg_json: the ORKG object to work on
        :param jsonld: the resulting JSON-LD part
        :param context: the accompanying context for JSON-LD
        :return: Nothing, since it operates on the parameter dictionaries
        """
        # add predicates into context
        for prop, url in [(p, self.PREDICATE_PREFIX + p) for p in orkg_json.keys()]:
            context[prop] = url

        # handle each predicate
        for key, entities in orkg_json.items():
            jsonld[key] = []
            for entity in entities:
                jsonld_entity = {}
                # handle dataframe values
                if '@df' in entity:
                    jsonld_entity = self.__handle_csvw_dataset(entity['@df'], context)
                # handle @id
                if '@id' in entity:
                    if '@type' in entity:  # TODO: replace with match in python 3.10
                        entity_type = entity['@type']
                        if entity_type == "resource":
                            jsonld_entity['@id'] = self.RESOURCE_PREFIX + entity['@id']
                        elif entity_type == "class":
                            jsonld_entity['@id'] = self.CLASS_PREFIX + entity['@id']
                        elif entity_type == "predicate":
                            jsonld_entity['@id'] = self.PREDICATE_PREFIX + entity['@id']
                        else:
                            # FIXME: the literal value should be looked up and added here properly
                            jsonld_entity['@literal'] = entity['@id']
                    else:
                        if entity['@id'][0] == "R":
                            jsonld_entity['@id'] = self.RESOURCE_PREFIX + entity['@id']
                        else:
                            # FIXME: the literal value should be looked up and added here properly
                            jsonld_entity['@literal'] = entity['@id']

                # handle text
                if 'text' in entity:
                    literal_value = entity['text']
                    if 'datatype' in entity:
                        literal_value = f"{literal_value}^^{jsonld_entity['datatype']}"
                    jsonld[key].append(literal_value)

                # handle label
                if 'label' in entity:
                    jsonld_entity['@id'] = self.get_blank_node_id()
                    jsonld_entity['label'] = entity['label']
                    if 'classes' in entity:
                        jsonld_entity['@type'] = [self.CLASS_PREFIX + c for c in entity['classes']]

                    # handle values
                    if 'values' in entity:
                        self.handle_values(entity["values"], jsonld_entity, context)
                # append newly created json-ld entity property
                if len(jsonld_entity) > 0:
                    jsonld[key].append(jsonld_entity)


if __name__ == '__main__':
    import json

    sample = """{"resource":{"name":"Statistically significant hypothesis test with IRE binding dependent variable on failing and non-failing hearts","classes":["C1007"],"values":{"P27011":[{"label":"http://purl.obolibrary.org/obo/GO_0030350","classes":["URI"]}],"P4003":[{"classes":["C1003"],"values":{"P9003":[{"classes":["C1004"],"values":{"P9004":[{"text":"1.3111247517411605e-08"}]},"label":""}]},"label":"the p-value of the statistical hypothesis test"}],"P4015":[{"label":"data.csv","classes":["C21033"]}]}}}"""
    print(JsonLDConverter("https://orkg.org/").convert_2_jsonld(json.loads(sample)))



from unittest import TestCase
from orkg import ORKG


class TestDummy(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """

    orkg = ORKG()

    def test_200_dummy(self):
        fake_content = {'orkg': 'super easy'}
        resp = self.orkg.dummy.create_200_response(content=fake_content)
        self.assertTrue(resp.succeeded)
        self.assertEqual(resp.content, fake_content)

    def test_404_dummy(self):
        fake_content = [{'msg': 'bad content'}]
        resp = self.orkg.dummy.create_404_response(content=fake_content)
        self.assertFalse(resp.succeeded)
        self.assertEqual(resp.content, fake_content)

    def test_generic_dummy(self):
        fake_content = {'error': 'server died'}
        fake_code = '500'
        resp = self.orkg.dummy.create_xxx_response(code=fake_code, content=fake_content)
        self.assertFalse(resp.succeeded)
        self.assertEqual(resp.status_code, fake_code)





from unittest import TestCase
from orkg import ORKG
from tempfile import NamedTemporaryFile


class TestPapers(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """
    orkg = ORKG()

    def test_add(self):
        paper = {"predicates": [], "paper": {
            "title": "Open Research Knowledge Graph: Next Generation Infrastructure for Semantic Scholarly Knowledge",
            "doi": "10.1145/3360901.3364435",
            "authors": [{"label": "Mohamad Yaser Jaradeh"}, {"label": "Allard Oelen"}, {"label": "Kheir Eddine Farfar"},
                        {"label": "Manuel Prinz"}, {"label": "Jennifer D'Souza"}, {"label": "Gábor Kismihók"},
                        {"label": "Markus Stocker"}, {"label": "Sören Auer"}], "publicationMonth": "",
            "publicationYear": 2019,
            "publishedIn": "Proceedings of the 10th International Conference on Knowledge Capture  - K-CAP '19",
            "researchField": "R11", "contributions": [{"name": "Contribution 1", "values": {"P32": [
                {"@temp": "_5254e420-ae9a-13ef-1c18-e716b9ea5c2b", "label": "Similarity measures", "class": "Problem",
                 "values": {}}], "P3": [
                {"text": "    Find similar research contributions inside the ORKG and suggest them to the user"}],
                "P1": [{
                    "@temp": "_851da0a6-7e83-c276-8df3-f66d0a680b30",
                    "label": "TF/iDF",
                    "values": {
                        "P2": [
                            {
                                "text": "Term frequency, inverse document frequency technique"
                            }
                        ]
                    }
                }
                ]
            }}]}}
        res = self.orkg.papers.add(params=paper)
        self.assertTrue(res.succeeded)

    def test_add_csv(self):
        csv_content = """paper:title,paper:authors,paper:doi,paper:publication_month,paper:publication_year,paper:research_field,contribution:research_problem,R0 estimates (average),95% Confidence interval,resource:Location,Study date,mean time interval between onset and hospital quarantine,Methods,mean incubation period,mean serial interval,study period name,Approaches,data collection tool,measurement condition 1,measurement condition 2,common symptoms,mean of serial interval,standard deviation of serial interval,code,incubation SD,incubation period mean,scenario,dataset description
Characterizing the transmission and identifying the control strategy for COVID-19 through epidemiological modeling,Ke K. Zhang; Linglin Xie; Lauren Lawless; Huijuan Zhou; Guannan Gao; Chengbin Xue,10.1101/2020.02.24.20026773,2,2020,R57,COVID-19 reproductive number,5.5,5.3-5.8,Mainland of China,,12.5 days,Monte carlo simulation,,,,,,,,,,,,,,,
Transmission interval estimates suggest pre-symptomatic spread of COVID-19,Lauren Tindale; Michelle Coombe; Jessica E Stockdale; Emma Garlock; Wing Yin Venus Lau; Manu Saraswat; Yen-Hsiang Brian Lee; Louxin Zhang; Dongxuan Chen; Jacco Wallinga; Caroline Colijn,10.1101/2020.03.03.20029983,3,2020,R57,COVID-19 reproductive number,1.97,1.45-2.48,Singapore,2020-01-19/2020-02-26,,,"7.1 (6.13, 8.25) days","4.56 (2.69, 6.42) days",,,,,,,,,,,,,
Transmission interval estimates suggest pre-symptomatic spread of COVID-19,Lauren Tindale; Michelle Coombe; Jessica E Stockdale; Emma Garlock; Wing Yin Venus Lau; Manu Saraswat; Yen-Hsiang Brian Lee; Louxin Zhang; Dongxuan Chen; Jacco Wallinga; Caroline Colijn,10.1101/2020.03.03.20029983,3,2020,R57,COVID-19 reproductive number,1.87,1.65-2.09,"Tianjin, China",2020-01-21/2020-02-27,,,"9 (7.92, 10.2) days","4.22 (3.43, 5.01) days",,,,,,,,,,,,,
"""
        csv_file = NamedTemporaryFile(mode="w", delete=False)
        csv_file.write(csv_content)
        csv_path = csv_file.name
        csv_file.close()

        res = self.orkg.papers.add_csv(file=csv_path)
        self.assertTrue(res.succeeded)
        self.assertTrue(len(res.content) == 3)



from unittest import TestCase
from orkg import ORKG


class TestStatements(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """
    orkg = ORKG()

    def test_by_id(self):
        res = self.orkg.statements.by_id('S1')
        self.assertTrue(res.succeeded)

    def test_get(self):
        res = self.orkg.statements.get(size=10)
        self.assertTrue(res.succeeded)
        self.assertEqual(len(res.content), 10)

    def test_get_with_items(self):
        count = 30
        res = self.orkg.statements.get(size=count)
        self.assertTrue(res.succeeded)
        self.assertEqual(len(res.content), count)

    def test_add(self):
        subject = self.orkg.resources.add(label="ORKG Subject").content['id']
        predicate = self.orkg.predicates.add(label="ORKG Predicate").content['id']
        object = self.orkg.literals.add(label="ORKG Object").content['id']
        res = self.orkg.statements.add(subject_id=subject, predicate_id=predicate, object_id=object)
        self.assertTrue(res.succeeded)

    def test_update(self):
        st_id = 'S1'
        new_p_id = 'P1'
        res = self.orkg.statements.update(id=st_id, predicate_id=new_p_id)
        self.assertTrue(res.succeeded)
        res = self.orkg.statements.by_id(st_id)
        self.assertTrue(res.succeeded)
        self.assertEqual(res.content['predicate']['id'], new_p_id)

    def test_exists(self):
        found = self.orkg.statements.exists(id='S1')
        self.assertTrue(found)

    def test_not_exists(self):
        found = self.orkg.statements.exists(id='SS1')
        self.assertFalse(found)

    def test_delete(self):
        subject = self.orkg.resources.add(label="ORKG Subject 1").content['id']
        predicate = self.orkg.predicates.add(label="ORKG Predicate 1").content['id']
        object = self.orkg.literals.add(label="ORKG Object 1").content['id']
        res = self.orkg.statements.add(subject_id=subject, predicate_id=predicate, object_id=object)
        st_id = res.content['id']
        self.assertTrue(res.succeeded)
        res = self.orkg.statements.delete(id=st_id)
        self.assertTrue(res.succeeded)
        found = self.orkg.statements.exists(id=st_id)
        self.assertFalse(found)



from unittest import TestCase
from orkg import ORKG
import numpy as np


class TestTemplates(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """
    orkg = ORKG(host="https://sandbox.orkg.org")

    def test_materialize(self):
        self.orkg.templates.materialize_templates(templates=['R12002'], verbose=False)
        print(self.orkg.templates.list_templates())
        self.assertTrue(True)

    def test_df_template(self):
        from pandas import DataFrame as df
        lst = ['this', 'is', 'fancy']
        lst2 = [4, 2, np.nan]
        param = df(list(zip(lst, lst2)), columns=['word', 'length'])
        self.orkg.templates.materialize_template(template_id='R199091')
        self.orkg.templates.test_df(label="what!", dataset=(param, 'Fancy Table'), uses_library="pyORKG").pretty_print(format='json-ld')
        self.assertTrue(True)

    def test_something(self):
        from math import ceil
        import pandas as pd
        from scipy.stats import ttest_ind
        self.orkg.templates.materialize_template('R12002')
        tp = self.orkg.templates

        df = pd.read_csv('data.csv')
        tt = ttest_ind(df['non-failing heart (NF)'],
                       df['failing heart (F)'],
                       equal_var=False, nan_policy='omit')
        pvalue = tt.pvalue
        pvalue_ceil = ceil(pvalue * 1000) / 1000.0
        pvalue_str = '{:.16f}'.format(pvalue)

        tp.students_ttest(
            label='Statistically significant hypothesis test with IRE binding dependent variable on failing and non-failing hearts (p<{})'.format(
                pvalue_ceil),
            has_dependent_variable='http://purl.obolibrary.org/obo/GO_0030350',
            # the study design dependent variable (iron-responsive element binding)
            has_specified_input=(
            df, 'Summary data showing iron-responsive element (IRE) binding activity in LV tissue samples'),
            # the input dataset
            has_specified_output=tp.pvalue('the p-value of the statistical hypothesis test (p<{})'.format(pvalue_ceil),
                                           tp.scalar_value_specification('{}'.format(pvalue_str), pvalue_str)
                                           ),
        ).serialize_to_file('article.contribution.1.json', format='json-ld')
        self.assertTrue(True)




from orkg.utils import NamespacedClient, query_params, dict_to_url_params, admin_functionality
from orkg.out import OrkgResponse
from pandas import read_csv
import os


class ResourcesClient(NamespacedClient):

    def by_id(self, id) -> OrkgResponse:
        """
        Lookup a resource by id
        :param id: the id of the resource to lookup
        :return: an OrkgResponse object containing the resource
        """
        self.client.backend._append_slash = True
        response = self.client.backend.resources(id).GET()
        return self.client.wrap_response(response)

    @query_params("q", "exact", "page", "size", "sort", "desc", "exclude")
    def get(self, params=None) -> OrkgResponse:
        """
        Fetch a list of resources, with the possibility to paginate the results and filter them out based on label
        :param q: search term of the label of the resource (optional)
        :param exact: whether to check for the exact search term or not (optional) -> bool
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :param exclude: classes to be excluded in search (optional)
        :return: an OrkgResponse object contains the list of resources
        """
        self.handle_sort_params(params)
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.resources.GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.resources.GET()
        return self.client.wrap_response(response)

    @query_params("id", "label", "classes")
    def add(self, params=None) -> OrkgResponse:
        """
        Create a new resource in the ORKG instance
        :param id: the specific id to add (optional)
        :param label: the label of the new resource (optional)
        :param classes: list of classes to assign the resource to (optional)
        :return: an OrkgResponse object containing the newly created resource
        """
        if len(params) == 0:
            raise ValueError("at least label should be provided")
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.resources.POST(json=params, headers=self.auth)
        return self.client.wrap_response(response)

    @query_params("id", "label", "classes")
    def find_or_add(self, params=None) -> OrkgResponse:
        """
        find or create a new resource in the ORKG instance
        :param id: the specific id to add (optional)
        :param label: the label of the new resource (optional)
        :param classes: list of classes to assign the resource to (optional)
        :return: an OrkgResponse object containing the found or newly created resource
        """
        if len(params) == 0:
            raise ValueError("at least label should be provided")
        else:
            if "id" in params:
                # check if a resource with this id is there
                found = self.by_id(params['id'])
                if found.succeeded:
                    return found
            # check if a resource with this label is there
            found = self.get(q=params['label'], exact=True, size=1)
            if found.succeeded:
                if isinstance(found.content, list) and len(found.content) > 0:
                    found.content = found.content[0]
                    return found
            # None found, what the hell! let's create a new one
            self.client.backend._append_slash = True
            response = self.client.backend.resources.POST(json=params, headers=self.auth)
        return self.client.wrap_response(response)

    @query_params("label", "classes")
    def update(self, id, params=None) -> OrkgResponse:
        """
        Update a resource with a specific id
        :param id: the id of the resource to update
        :param label: the new label (optional)
        :param classes: the updated list of classes (optional)
        :return: an OrkgResponse object contains the newly updated resource
        """
        if len(params) == 0:
            raise ValueError("at least label should be provided")
        else:
            if not self.exists(id):
                raise ValueError("the provided id is not in the graph")
            self.client.backend._append_slash = True
            response = self.client.backend.resources(id).PUT(json=params, headers=self.auth)
        return self.client.wrap_response(response)

    @admin_functionality
    def update_observatory(self, id: str, observatory_id: str, organization_id: str) -> OrkgResponse:
        """
        Update the observatory and the organization of a particular resource
        NOTE: this is a call only for admins
        :param id: The id of the resource to update
        :param observatory_id: the new observatory id
        :param organization_id: the new organization id
        :return: an OrkgResponse object contains the newly updated resource
        """
        if not self.exists(id):
            raise ValueError("the provided id is not in the graph")
        self.client.backend._append_slash = False
        content = {
            'observatory_id': observatory_id,
            'organization_id': organization_id
        }
        response = self.client.backend.resources(id).observatory.PUT(json=content, headers=self.auth)
        return self.client.wrap_response(response)

    def exists(self, id) -> bool:
        """
        Check if a resource exists in the graph
        :param id: the id of the resource to check
        :return: true if found, otherwise false
        """
        return self.by_id(id).succeeded

    @admin_functionality
    def delete(self, id) -> OrkgResponse:
        if not self.exists(id):
            raise ValueError("the provided id is not in the graph")
        self.client.backend._append_slash = True
        response = self.client.backend.resources(id).DELETE(headers=self.auth)
        return self.client.wrap_response(response)

    def save_dataset(self, file, label, dimensions) -> OrkgResponse:
        """
        Create a resource of a tabular data using RDF Datacube vocabulary starting from a CSV file
        :param file: CSV file containing the table
        :param label: label of the resource
        :param dimensions: a list of column(s) name(s) that represent the dimensions.
        :return: the resource ID of the dataset
        """
        
        if not file:
            raise ValueError("the file should be provided")
        if not label:
            raise ValueError("the label should be provided")

        if not os.path.exists(file):
            raise ValueError(f"file path doesn't exist!, path={file}")
                                
        dataset = read_csv(file, dtype=str)                      
        # Vocabulary Classes
        cQB_DATASET_CLASS = self.client.classes.find_or_add(id="QBDataset", label="qb:DataSet").content['id']
        cDataStructureDefinition = self.client.classes.find_or_add(label="qb:DataStructureDefinition").content['id']
        cComponentSpecification = self.client.classes.find_or_add(label="qb:ComponentSpecification").content['id']
        cComponentProperty = self.client.classes.find_or_add(label="qb:ComponentProperty").content['id']
        cDimensionProperty = self.client.classes.find_or_add(label="qb:DimensionProperty").content['id']
        cMeasureProperty = self.client.classes.find_or_add(label="qb:MeasureProperty").content['id']
        cObservation = self.client.classes.find_or_add(label="qb:Observation").content['id']
        # Vocabulary Predicates
        pStructure = self.client.predicates.find_or_add(label="structure").content['id']
        pComponent = self.client.predicates.find_or_add(label="component").content['id']
        pDimension = self.client.predicates.find_or_add(label="dimension").content['id']
        pMeasure = self.client.predicates.find_or_add(label="measure").content['id']
        pDataSet = self.client.predicates.find_or_add(label="dataSet").content['id']
        pOrder = self.client.predicates.find_or_add(label="order").content['id']
        # Create dataset resource
        rds = self.client.resources.add(label=label, classes=[cQB_DATASET_CLASS])
        rds_id = rds.content['id']
        # Create Data Structure Definition
        rDsd = self.client.resources.add(label="Data Structure Definition", classes=[cDataStructureDefinition]).content['id']
        self.client.statements.add(subject_id=rds_id, predicate_id=pStructure, object_id=rDsd)
        # Schema Definition
        cs = dict()  # component specifications
        dt = dict()  # component properties
        dtP = dict()  # resources used in the predicate position
        for index, column in enumerate(dataset.columns, start=1):
            cs[column] = self.client.resources.add(label="Component Specification " + column, classes=[cComponentSpecification]).content['id']
            self.client.statements.add(subject_id=rDsd, predicate_id=pComponent, object_id=cs[column])
            # Order
            lOrder = self.client.literals.add(label=str(index)).content['id']
            self.client.statements.add(subject_id=cs[column], predicate_id=pOrder, object_id=lOrder)
            if column in dimensions:
                dt[column] = self.client.resources.add(label=column, classes=[cComponentProperty, cDimensionProperty]).content['id']
                self.client.statements.add(subject_id=cs[column], predicate_id=pDimension, object_id=dt[column])
            else:
                dt[column] = self.client.resources.add(label=column, classes=[cComponentProperty, cMeasureProperty]).content['id']
                self.client.statements.add(subject_id=cs[column], predicate_id=pMeasure, object_id=dt[column])
            # Name a predicate with the resource ID
            dtP[column] = self.client.predicates.find_or_add(label=dt[column]).content['id']
            
        # Observations
        for index, row in dataset.iterrows():
            ro = self.client.resources.add(label='Observation #{}'.format(index+1), classes=[cObservation]).content['id']
            self.client.statements.add(subject_id=ro, predicate_id=pDataSet, object_id=rds_id)
            # set the values
            for column in dataset.columns:
                lid = self.client.literals.add(label=str(row[column])).content['id']
                self.client.statements.add(subject_id=ro, predicate_id=dtP[column], object_id=lid)
        return rds




from orkg.utils import NamespacedClient, query_params
from orkg.out import OrkgResponse
from pandas import read_csv
import re
from ast import literal_eval
import os


class PapersClient(NamespacedClient):

    def add(self, params=None, merge_if_exists=False) -> OrkgResponse:
        """
        Create a new paper in the ORKG instance
        :param params: paper Object
        :param merge_if_exists: merge the papers if they exist in the graph and append contributions
        :return: an OrkgResponse object containing the newly created paper resource
        """
        self.client.backend._append_slash = True
        response = self.client.backend.papers.POST(json=params, params={'mergeIfExists': merge_if_exists}, headers=self.auth)
        return self.client.wrap_response(response)

    @query_params("file", "standard_statements")
    def add_csv(self, params=None) -> OrkgResponse:
        """
        Create a new paper in the ORKG instance
        :param file: csv file containing the papers, with at least a column with title: "paper:title"
        :param standard_statements: dict with default statements, should contain CSV_PLACEHOLDER (optional)
        :return: an list with paper ids and contribution ids 
        """
        if 'file' not in params:
            return
        if not os.path.exists(params['file']):
            raise ValueError(f"file path doesn't exist!, path={params['file']}")

        df = read_csv(params['file'], dtype=str)
        contribution_ids = []

        for index in range(len(df)):
            paper = df.iloc[index].to_dict()

            # at least a paper title should be present
            if 'paper:title' not in paper: # TODO: enable crossref lookup based on DOI
                print('Skipping paper, column "paper:title" not found')
                continue

            # check for paper metadata
            title = paper['paper:title']
            authors = [{'label': author.strip()} for author in str(paper['paper:authors']).split(';')] if 'paper:authors' in paper and paper['paper:authors'] == paper['paper:authors'] else []
            publication_month = self.__setValue('paper:publication_month', paper, 1)
            publication_year = self.__setValue('paper:publication_year', paper, 2000)
            research_field = self.__setValue('paper:research_field', paper, 'R11')
            doi = self.__setValue('paper:doi', paper, '')
            url = ''
            published_in = ''
            standard_statements = literal_eval(
                params['standard_statements']) if 'standard_statements' in params else {}
            contribution_statements = {}

            # remove metadata from paper dict, already added above
            metadata_headers = ['paper:title', 'paper:authors', 'paper:publication_month', 'paper:publication_year', 'paper:doi', 'paper:research_field']
            [paper.pop(key) if key in paper else paper for key in metadata_headers] 

            research_problems = []    
            # generate the statements based on the remaining columns
            for predicate in paper:
                value = paper[predicate]

                # add research problem (one or more)
                if (predicate.startswith('contribution:research_problem')):
                    research_problem = self.__setValue(predicate, paper, '')

                    if (research_problem != ''):
                        research_problems.append(research_problem)

                    continue


                # filter out nan values
                if value != value:
                    continue
                
                predicate = re.sub('\.[1-9]+$', '', predicate) # to make columns unique, pandas appends a dot and number to duplicate columns, remove this here
                
                value_is_resource = False

                # if predicate starts with resource:, insert it as resource instead of literal
                if predicate.startswith('resource:'):
                    value_is_resource = True
                    predicate = predicate[len('resource:'):]

                predicate_id = self.client.predicates.find_or_add(label=predicate).content['id']

                if not value_is_resource:
                    if predicate_id in contribution_statements:
                        contribution_statements[predicate_id].append({'text': value})
                    else:
                        contribution_statements[predicate_id] = [
                            {'text': value}
                        ]
                else:
                    resource_id = self.client.resources.find_or_add(label=value).content['id']
                    if predicate_id in contribution_statements:
                        contribution_statements[predicate_id].append({'@id': resource_id})
                    else:
                        contribution_statements[predicate_id] = [
                            {'@id': resource_id}
                        ]

            # add default statements if they are present 
            # TODO: only works for one level now, make it recursive
            statements_to_insert = standard_statements.copy()

            if len(statements_to_insert) > 0:
                for predicate in statements_to_insert:
                    if isinstance(statements_to_insert[predicate], list):  # if is array
                        for i in range(len(statements_to_insert[predicate])):
                            if statements_to_insert[predicate][i]['values'] == 'CSV_PLACEHOLDER':
                                statements_to_insert[predicate][i]['values'] = contribution_statements

                    if not re.search("^[P]+[a-zA-Z0-9]*$", predicate):
                        predicate_id = self.client.predicates.find_or_add(label=predicate).content['id']
                        statements_to_insert[predicate_id] = statements_to_insert[predicate]
                        del statements_to_insert[predicate]

                contribution_statements = statements_to_insert

            # add research problem to the contribution 
            if len(research_problems) > 0:
                contribution_statements['P32'] = []
                
                for research_problem in research_problems:
                    research_problem_id = \
                        self.client.resources.find_or_add(label=research_problem, classes=['Problem']).content['id']

                    # P32 has research problem 
                    contribution_statements['P32'].append({'@id': research_problem_id})

            # check if a paper with this title already exists 
            existing_papers = self.client.resources.get(q=title, exact=True).content
            existing_paper_id = 0

            for existing_paper in existing_papers:
                if 'Paper' in existing_paper['classes']:
                    existing_paper_id = existing_paper['id']
                    break

            # paper does not yet exist
            if existing_paper_id == 0:
                paper = {
                    'paper': {
                        'title': title,
                        'authors': authors,
                        'publicationMonth': publication_month,
                        'publicationYear': publication_year,
                        'researchField': research_field,
                        'doi': doi,
                        'url': url,
                        'publishedIn': published_in,
                        'contributions': [{
                            'name': 'Contribution 1',
                            'values': contribution_statements
                        }]
                    }
                }
                response = self.add(paper)

                if 'id' in response.content:
                    paper_id = response.content['id']
                    paper_statements = self.client.statements.get_by_subject(subject_id=paper_id, size=10000).content
                    for statement in paper_statements:
                        if statement['predicate']['id'] == 'P31':
                            contribution_ids.append(statement['object']['id'])
                            print('Added paper:', str(paper_id))
                else:
                    print('Error adding paper: ', str(response.content))

            # paper exists already, so add a contribution       
            else:
                paper_statements = self.client.statements.get_by_subject(subject_id=existing_paper_id, size=10000).content
                contribution_amount = 0

                for paper_statement in paper_statements:
                    if paper_statement['predicate']['id'] == 'P31':
                        contribution_amount += 1

                contribution_id = self.client.resources.add(label='Contribution ' + str(contribution_amount + 1),
                                                            classes=['Contribution']).content['id']

                self.client.statements.add(subject_id=existing_paper_id, predicate_id='P31', object_id=contribution_id)
                self.__create_statements(contribution_id, contribution_statements)

                contribution_ids.append(contribution_id)

                print('Added contribution:', str(contribution_id), 'to paper:', str(existing_paper_id))

        return self.client.wrap_response(content=contribution_ids, status_code="201")
    
    # function to check if key exists and value is not NaN
    def __setValue(self, key, dict, defaultValue):
        return dict[key] if key in dict and dict[key] == dict[key] else defaultValue

    def __create_statements(self, subject_id, statements):
        for predicate_id in statements:
            values = statements[predicate_id]

            for value in values:
                if 'text' in value:
                    literal_id = self.client.literals.add(label=value['text']).content['id']
                    self.client.statements.add(subject_id=subject_id, predicate_id=predicate_id, object_id=literal_id)
                elif '@id' in value:
                    self.client.statements.add(subject_id=subject_id, predicate_id=predicate_id, object_id=value['@id'])
                elif 'label' in value:
                    resource_id = self.client.resources.add(label=value['label']).content['id']
                    self.client.statements.add(subject_id=subject_id, predicate_id=predicate_id, object_id=resource_id)
                    self.__create_statements(resource_id, value['values'])



import networkx as nx

from typing import List, Union

from orkg import ORKG


def subgraph(client: ORKG, thing_id: str, blacklist: Union[str, List[str]] = '', max_level: int = -1) -> nx.DiGraph:
    """
    Obtains a networkx directed graph representation of any ORKG component given by its thing_id.
    E.g. of ORKG components: Paper, Contribution, Comparison, Template

    It starts from the thing_id resource and traverses the graph until all literals are reached.

    :param client: orkg.ORKG client used to connect with ORKG backend.
    :param thing_id: Any subject, object or predicate ID in the ORKG.
    :param blacklist: Class(es) to be excluded from the subgraph. E.g. 'ResearchField'
        (see `orkgc:ResearchField <https://orkg.org/class/ResearchField>`_).
        Note that the first subgraph level will always be included.
    :param max_level: Deepest subgraph's level to traverse.
    """
    blacklist = [blacklist] if isinstance(blacklist, str) else blacklist

    response = client.statements.bundle(thing_id, params={
        'includeFirst': 'true',
        'blacklist': ','.join(blacklist),
        'maxLevel': max_level
    })

    if not response.succeeded:
        raise ValueError('Something went wrong while connecting to ORKG backend with host {}'.format(client.host))

    statements = response.content['statements']

    if not statements:
        raise ValueError('Nothing found for the provided ID: {}'.format(thing_id))

    return _construct_subgraph(nx.DiGraph(), statements)


def _construct_subgraph(subgraph: nx.DiGraph, statements: list) -> nx.DiGraph:
    """
    Constructs a subgraph represented by the given RDF ``statements``.

    :param subgraph: Initial networkx.DiGraph to extend with nodes and edges.
    :param statements: List of all RDF statements describing the subgraph.
    """
    for statement in statements:
        # NetworkX does not create a node or an edge double, therefore, the following implementation works :)
        start_node = _create_node_from_thing(subgraph, statement['subject'])
        target_node = _create_node_from_thing(subgraph, statement['object'])
        _create_edge(subgraph, start_node, target_node, statement['predicate'])

    return subgraph


def _create_node_from_thing(subgraph: nx.DiGraph, thing: dict) -> str:
    subgraph.add_node(
        node_for_adding=thing['id'],
        id=thing['id'],
        label=thing['label'],
        class_=thing['_class'],
        classes=thing.get('classes', []),
        description=thing.get('description', ''),
        datatype=thing.get('datatype', '')
    )
    return thing['id']


def _create_edge(subgraph: nx.DiGraph, root_node: str, target_node: str, predicate) -> None:
    subgraph.add_edge(
        root_node,
        target_node,
        id=predicate['id'],
        label=predicate['label'],
        class_=predicate['_class'],
    )



from unittest import TestCase
from orkg import ORKG


class TestPredicates(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """
    orkg = ORKG()

    def test_by_id(self):
        res = self.orkg.predicates.by_id('P1')
        self.assertTrue(res.succeeded)

    def test_get(self):
        size = 10
        res = self.orkg.predicates.get(size=size)
        self.assertTrue(res.succeeded)
        self.assertEqual(len(res.content), size)

    def test_get_with_items(self):
        count = 30
        res = self.orkg.predicates.get(size=count)
        self.assertTrue(res.succeeded)
        self.assertEqual(len(res.content), count)

    def test_add(self):
        label = "test predicate"
        res = self.orkg.predicates.add(label=label)
        self.assertTrue(res.succeeded)
        self.assertEqual(res.content['label'], label)

    def test_find_or_add(self):
        import random
        import string
        label = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(15))
        old = self.orkg.predicates.add(label=label)
        self.assertTrue(old.succeeded, 'Creating first predicate is a success')
        self.assertEqual(old.content['label'], label, 'The first predicate has the correct label')
        new = self.orkg.predicates.find_or_add(label=label)
        self.assertTrue(new.succeeded, 'Creating second predicate is a success')
        self.assertEqual(new.content['id'], old.content['id'], 'the two predicates have the same id')

    def test_update(self):
        res = self.orkg.predicates.add(label="Coco predicate")
        self.assertTrue(res.succeeded)
        label = "Test predicate"
        res = self.orkg.predicates.update(id=res.content['id'], label=label)
        self.assertTrue(res.succeeded)
        res = self.orkg.predicates.by_id(res.content['id'])
        self.assertTrue(res.succeeded)
        self.assertEqual(res.content['label'], label)



from unittest import TestCase
from orkg import ORKG


class TestLiterals(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """
    orkg = ORKG()

    def test_by_id(self):
        res = self.orkg.literals.by_id('L1')
        self.assertTrue(res.succeeded)

    def test_get(self):
        res = self.orkg.literals.get_all()
        self.assertTrue(res.succeeded)

    def test_get_term(self):
        term = 'Python'
        res = self.orkg.literals.get_all(q=term)
        self.assertTrue(res.succeeded)

    def test_add(self):
        label = "ORKG rocks"
        res = self.orkg.literals.add(label=label)
        self.assertTrue(res.succeeded)
        self.assertEqual(res.content['label'], label)

    def test_update(self):
        res = self.orkg.literals.add(label="ORKG")
        self.assertTrue(res.succeeded)
        label = "Open Research Knowledge Graph"
        res = self.orkg.literals.update(id=res.content['id'], label=label)
        self.assertTrue(res.succeeded)
        res = self.orkg.literals.by_id(res.content['id'])
        self.assertTrue(res.succeeded)
        self.assertEqual(res.content['label'], label)



from typing import Dict, List, Union
from orkg.utils import NamespacedClient, simcomp_available
from orkg.out import OrkgResponse


class JSONClient(NamespacedClient):

    @simcomp_available
    def save_json(self, resource_id: str, json_object: Union[Dict, List]) -> OrkgResponse:
        return self.client.wrap_response(
            self.client.simcomp.visualization.POST(
                json={"resourceId": resource_id, "jsonData": json_object}
            )
        )

    @simcomp_available
    def get_json(self, resource_id: str) -> OrkgResponse:
        params = f'?resourceId={resource_id}'
        return self.client.wrap_response(self.client.simcomp.visualization.GET(params))




from setuptools import find_packages
from distutils.core import setup

long_description = open('README.md').read()

setup(
  name='orkg',
  packages=find_packages(),
  version='0.14.6',
  license='MIT',
  description='Python client for the Open Research Knowledge Graph (ORKG) API',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author='Yaser Jaradeh',
  author_email='jaradeh@l3s.de',
  url='http://orkg.org/about',
  download_url='https://gitlab.com/TIBHannover/orkg/orkg-pypi/-/archive/0.14.6/orkg-pypi-0.14.6.tar.gz',
  keywords=['ORKG', 'Scholarly communication', 'API wrapper'],
  install_requires=[
          'hammock',
          'pandas',
          'tqdm',
          'nose',
          'networkx'
      ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8"
  ],
  test_suite='nose.collector',
  tests_require=['nose'],
)



from unittest import TestCase
from orkg import ORKG


class TestStats(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """
    orkg = ORKG()

    def test(self):
        res = self.orkg.stats.get()
        self.assertTrue(res.succeeded)
